/* $XConsortium$ */
/*
 * Adapted for XFree86 4.0.1 by Joe Krahn <krahn@niehs.nih.gov>
 *
 * Copyright 1998 by Francois-Regis Colin, <fcolin@cenatls.cena.dgac.fr> and
 *                  Frederic Lepied, <Fredric.Lepied@sugix.frmug.org>. France.
 *                                                                            
 * Permission  to use, copy,  modify, distribute, and  sell this software
 * and its  documentation for any purpose  is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that
 * both  that  copyright notice   and this  permission notice  appear  in
 * supporting documentation, and  that the names of Francois-Regis  Colin
 * and Frederic Lepied not be used in advertising or publicity pertaining
 * to distribution of     the software without specific,  written   prior
 * permission.   Francois-Regis   Colin   and  Frederic  Lepied make   no
 * representations  about the suitability   of   this software for    any
 * purpose.  It is provided "as is" without express or implied warranty.
 *                                                                            
 * FRANCOIS-REGIS COLIN AND FREDERIC LEPIED  DISCLAIM ALL WARRANTIES WITH
 * REGARD  TO THIS   SOFTWARE,  INCLUDING   ALL IMPLIED  WARRANTIES    OF
 * MERCHANTABILITY  AND  FITNESS, IN NO EVENT    SHALL FREDERIC LEPIED BE
 * LIABLE  FOR  ANY SPECIAL,  INDIRECT OR   CONSEQUENTIAL DAMAGES  OR ANY
 * DAMAGES  WHATSOEVER   RESULTING FROM LOSS  OF   USE, DATA  OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT  OF OR IN CONNECTION  WITH THE USE  OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 */

/* $XFree86: xc/programs/Xserver/hw/xfree86/input/dialbox/dialbox.c,v *******/

/*
 * This driver handles SGI dial and button boxes protocol.
 * These are actually two devices that work together. Most
 * people use just the dialbox, so the driver will be called "dialbox"
 * even though it supports the button box as well.
 */
/*****************************************************************************
 *	Standard Headers
 ****************************************************************************/

#include <misc.h>
#include <xf86.h>
#define NEED_XF86_TYPES
#include <xf86_ansic.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <xisb.h>
#include <exevents.h>

/*****************************************************************************
 *	Local Headers
 ****************************************************************************/
#include "dialbox.h"

/*****************************************************************************
 *	Variables without includable headers
 ****************************************************************************/

/*****************************************************************************
 *	Local Variables
 ****************************************************************************/

static InputInfoPtr DialPreInit (InputDriverPtr drv, IDevPtr dev, int flags);

static InputDriverRec DIALBOX = {
  1,				/* driver version */
  "dialbox",			/* driver name */
  NULL,				/* identify */
  DialPreInit,			/* pre-init */
  NULL,				/* un-init */
  NULL,				/* module */
  0				/* ref count */
};

#ifdef XFree86LOADER		/* Do we really still need this part in an #ifdef ? */

#define wait_for_fd(fd) xf86WaitForInput((fd), 1000)

XF86ModuleVersionInfo dialVersionRec = {
  "dials+buttons",
  /*"dialbox", */
  MODULEVENDORSTRING,
  MODINFOSTRING1,
  MODINFOSTRING2,
  XF86_VERSION_CURRENT,
  1, 0, 0,
  ABI_CLASS_XINPUT,
  ABI_XINPUT_VERSION,
  MOD_CLASS_XINPUT,
  {0, 0, 0, 0}
};

/* This is not an accurate list; it was borrowed from the microtouch driver */
static const char *reqSymbols[] = {
  "AddEnabledDevice",
  "ErrorF",
  "InitButtonClassDeviceStruct",
  "InitProximityClassDeviceStruct",
  "InitValuatorAxisStruct",
  "InitValuatorClassDeviceStruct",
  "RemoveEnabledDevice",
  "Xcalloc",
  "Xfree",
  "XisbBlockDuration",
  "XisbFree",
  "XisbNew",
  "XisbRead",
  "XisbTrace",
  "XisbWrite",
  "screenInfo",
  "xf86AddInputDriver",
  "xf86AllocateInput",
  "xf86CloseSerial",
  "xf86CollectInputOptions",
  "xf86ErrorF",
  "xf86ErrorFVerb",
  "xf86FindOptionValue",
  "xf86FlushInput",
  "xf86GetMotionEvents",
  "xf86GetVerbosity",
  "xf86LoaderReqSymLists",
  "xf86MotionHistoryAllocate",
  "xf86Msg",
  "xf86NameCmp",
  "xf86OpenSerial",
  "xf86OptionListCreate",
  "xf86OptionListFree",
  "xf86OptionListReport",
  "xf86PostButtonEvent",
  "xf86PostMotionEvent",
  "xf86PostProximityEvent",
  "xf86ProcessCommonOptions",
  "xf86ScaleAxis",
  "xf86SetIntOption",
  "xf86SetSerial",
  "xf86SetStrOption",
  "xf86XInputSetScreen",
  "xf86XInputSetSendCoreEvents",
  "xf86memset",
  "xf86sscanf",
  "xf86strcmp",
  "xf86strlen",
  "xf86strncmp",
  "xf86strncpy",
  NULL
};

XF86ModuleData dialboxModuleData =
  { &dialVersionRec, DialSetupProc, DialTearDownProc };

static pointer
DialSetupProc (pointer module, pointer options, int *errmaj, int *errmin)
{
  xf86LoaderReqSymLists (reqSymbols, NULL);
  xf86AddInputDriver (&DIALBOX, module, 0);
  return (pointer) 1;
}

static void
DialTearDownProc (pointer p)
{
  LocalDevicePtr local = (LocalDevicePtr) p;
  DialPtr dial = (DialPtr) local->private;

  xf86RemoveEnabledDevice (local);
  /*dial_clear_leds (dial); */
  xf86CloseSerial (local->fd);
  XisbFree (dial->buffer);
  xfree (dial);
  xfree (local->name);
  xfree (local);
}

#endif /* XFree86LOADER */


static const char *default_options[] = {
  "Device", "/dev/dialbox",
  "BaudRate", "9600",
  "StopBits", "1",
  "DataBits", "8",
  "Parity", "None",
  "Vmin", "4",
  "Vtime", "1",
  "FlowControl", "None",
  NULL
};

/*****************************************************************************
 *	XFree86 Function Definitions
 ****************************************************************************/


/*
 * Was DialSetupProc, now it is the Pre-Init, no docs to tell me why yet.
 * We can no longer use errmaj/errmin to report error types.
 */
static InputInfoPtr
DialPreInit (InputDriverPtr drv, IDevPtr dev, int flags)
{
  LocalDevicePtr local;
  DialPtr dial;

  xf86ErrorFVerb (9, "Dialbox: Running Pre-Init.\n");

  dial = xcalloc (1, sizeof (DialRec));
  if (!dial)
    return NULL;

  local = xf86AllocateInput (drv, 0);
  if (!local)
    {
      xfree (dial);
      return NULL;
    }

  local->type_name = XI_KNOB_BOX;
  local->device_control = DialDeviceControl;
  local->read_input = DialReadInput;
  local->control_proc = DialControlProc;
  local->close_proc = DialCloseProc;

  local->switch_mode = DialSwitchMode;
  local->conversion_proc = DialConvert;
  local->reverse_conversion_proc = DialReverseConvert;

  local->dev = NULL;
  local->private = dial;
  local->private_flags = 0;
  local->flags = 0;
  local->conf_idev = dev;
/*
  local->fd = -1;
  local->atom = 0;
  local->history_size = 0;
  local->old_x = -1;
  local->old_y = -1;
*/

  xf86CollectInputOptions (local, default_options, NULL);

  xf86OptionListReport (local->options);

  xf86ErrorFVerb (4, "Dialbox Device = %s\n",
		  xf86SetStrOption (local->options, "Device", NULL));

  local->fd = xf86OpenSerial (local->options);
  if (local->fd == -1)
    {
      ErrorF ("DialBox driver unable to open serial port\n");
      goto PreInit_fail;
    }
  xf86ErrorFVerb (6, "DialBox: serial port opened successfully\n");

  /* Initialize other dial variabls from the XF86Config here */
  /* with:      dial->val = xf86SetIntOption( local->options, "Value", 0 ); */

  dial->flags = DIAL_FLAG_ABSOLUTE;
  dial->buffer = XisbNew (local->fd, DIAL_BUFFER_SIZE);

  DBG (9, XisbTrace (dial->buffer, 1));
  /*xf86ErrorFVerb(9, XisbTrace (dial->buffer, 1)); */

  /* 
   * Verify that hardware is attached and fuctional
   */

  if (DialQueryHardware (local) != Success)
    {
      ErrorF ("Unable to query/initialize DialBox hardware.\n");
      goto PreInit_fail;
    }

  if ( xf86SetBoolOption (local->options, "EnableButtons", TRUE) )
    {
      dial->flags |= DIAL_FLAG_ENABLE_BUTTONS;
      /* this results in an xstrdup that must be freed later */
      local->name =
	xf86SetStrOption (local->options, "DeviceName", "dial+buttons");
    }
  else
    {
      /* this results in an xstrdup that must be freed later */
      local->name =
	xf86SetStrOption (local->options, "DeviceName", "dialbox");
    }


  xf86ProcessCommonOptions (local, local->options);
  local->flags |= XI86_CONFIGURED;
  ErrorF ("Dial box initialization complete.\n");

  if (local->fd != -1)
    {
      RemoveEnabledDevice (local->fd);
      if (dial->buffer)
	{
	  XisbFree (dial->buffer);
	  dial->buffer = NULL;
	}
      xf86CloseSerial (local->fd);
    }
  RemoveEnabledDevice (local->fd);
  local->fd = -1;

  xf86ErrorFVerb (9, "Dialbox: Pre-Init Successful.\n");
  /* return the LocalDevice */
  return (local);

  /* 
   * If something went wrong, cleanup and return NULL
   */
PreInit_fail:
  if ((local) && (local->fd))
    xf86CloseSerial (local->fd);
  if ((local) && (local->name))
    xfree (local->name);
  if (local)
    xfree (local);

  if ((dial) && (dial->buffer))
    XisbFree (dial->buffer);
  if (dial)
    xfree (dial);

  xf86ErrorFVerb (9, "Dialbox: Pre-Init Failed.\n");
  return (NULL);
}


/******************************************************************************
 * dial box communication
 *****************************************************************************/
static int
dial_enable_button (DialRec * dial, int btns, int value)
{
  int i;
  unsigned char dcmd[5];

  unsigned char mask[4] = { 0, 0, 0, 0 };

  dial->bactive[btns] = value;

  for (i = 0; i < DIAL_NUM_BUTTONS; i++)
    {
      if (dial->bactive[i])
	SET (i, mask);
    }

  dcmd[0] = DIAL_SET_AUTO_MOM_BUTTONS;
  dcmd[1] = mask[0];
  dcmd[2] = mask[1];
  dcmd[3] = mask[2];
  dcmd[4] = mask[3];

  XisbWrite (dial->buffer, dcmd, sizeof (dcmd));

  dcmd[0] = DIAL_SET_BUTTONS_MOM_TYPE;
  dcmd[1] = mask[0];
  dcmd[2] = mask[1];
  dcmd[3] = mask[2];
  dcmd[4] = mask[3];

  XisbWrite (dial->buffer, dcmd, sizeof (dcmd));

  return 1;
}

/*
static int
dial_switch_leds (DialRec * dial)
{
  char cmd = DIAL_SET_ALL_LEDS;

  if (dial->buffer)
  return XisbWrite (dial->buffer, &cmd, 1);
  return !Success;
}

static int
dial_clear_leds (DialRec * dial)
{
  char cmd = DIAL_CLEAR_ALL_LEDS;
  if (dial->buffer)
    return XisbWrite (dial->buffer, &cmd, 1);
  return !Success;
}
*/

/* Enable valuator (val) IF value!=0 */
static int
dial_enable_valuator (DialRec * dial, int val, int value)
{
  int i;
  unsigned char dcmd[3];
  unsigned char mask[2] = { 0, 0 };

  dial->vactive[val] = value;

  xf86ErrorFVerb (9, "Dialbox: %sabling valuator %d\n", value ? "En" : "Dis",
		  val);

  for (i = 0; i < DIAL_NUM_VALUATORS; i++)
    {
      if (dial->vactive[i])
	SET (i, mask);
    }
  dcmd[0] = DIAL_SET_AUTO_DIALS;
#ifdef NOTDEF
  dcmd[1] = mask[0];
  dcmd[2] = mask[1];
#else
  dcmd[1] = 0xff;
  dcmd[2] = 0xff;
#endif
  XisbWrite (dial->buffer, dcmd, sizeof (dcmd));

  return 1;
}

static int
dial_set_resolution (DialRec * dial, int valNum, int valFilter)
{
  char dcmd[3];
  unsigned char filter = valFilter;


  dcmd[0] = DIAL_SET_FILTER;
  dcmd[1] = valNum;
  dcmd[2] = filter;

  XisbWrite (dial->buffer, dcmd, sizeof (dcmd));

  return 1;
}

#if 0
static int
dial_set_led (DialRec * dial, int led, int value)
{
  int i;
  char dcmd[5];
  unsigned char mask[4] = { 0, 0, 0, 0 };

  dial->ledstate[led] = value;

  for (i = 0; i < DIAL_NUM_LEDS; i++)
    {
      if (dial->ledstate[i])
	SET (i, mask);
    }
  dcmd[0] = DIAL_SET_LEDS;
  dcmd[1] = mask[0];
  dcmd[2] = mask[1];
  dcmd[3] = mask[2];
  dcmd[4] = mask[3];
  XisbWrite (dial->buffer, dcmd, sizeof (dcmd));

  return 1;
}

static int
dial_set_led_mask (DialRec * dial, unsigned int mask)
{
  unsigned char dcmd[5];

  dcmd[0] = DIAL_SET_LEDS;
  dcmd[1] = (unsigned char) (0x000000ff & mask);
  dcmd[2] = (unsigned char) ((0x0000ff00 & mask) >> 8);
  dcmd[3] = (unsigned char) ((0x00ff0000 & mask) >> 16);
  dcmd[4] = (unsigned char) ((0xff000000 & mask) >> 24);

  XisbWrite (dial->buffer, dcmd, sizeof (dcmd));

  return 1;
}
#endif

static void
dial_setup (DialRec * dial)
{
  int i;

  dial->ready = 1;
/*
  dial_switch_leds (dial);
  sleep (1);
  dial_clear_leds (dial);

  for (i = 0; i < DIAL_NUM_LEDS; i++)
    {
      dial_set_led (dial, i, dial->ledstate[i]);
    }
*/

  if (dial->flags & DIAL_FLAG_ENABLE_BUTTONS)
    {
      for (i = 0; i < DIAL_NUM_BUTTONS; i++)
	{
	  dial->bactive[i] = 1;
	  dial_enable_button (dial, i, dial->bactive[i]);
	}
    }


  for (i = 0; i < DIAL_NUM_VALUATORS; i++)
    {
      dial_set_resolution (dial, i, 0);
      dial_enable_valuator (dial, i, dial->vactive[i]);
    }

  dial->errorCount = 0;
  return;
}

#if 0
/*
 ***************************************************************************
 *
 * DialConfig --
 *	Configure the device.
 *
 ***************************************************************************
 */
#define PORT		1
#define DEVICENAME	2
#define THE_MODE	3
#define DEBUG_LEVEL     4
#define HISTORY_SIZE	5
#define ALWAYS_CORE	6

#define SECTION_NAME "dialbox"

static SymTabRec DialTab[] = {
  {ENDSUBSECTION, "endsubsection"},
  {PORT, "port"},
  {DEVICENAME, "devicename"},
  {THE_MODE, "mode"},
  {DEBUG_LEVEL, "debuglevel"},
  {HISTORY_SIZE, "historysize"},
  {ALWAYS_CORE, "alwayscore"},
  {-1, ""}
};

#define RELATIVE	1
#define ABSOLUTE	2

static SymTabRec ModeTabRec[] = {
  {RELATIVE, "relative"},
  {ABSOLUTE, "absolute"},
  {-1, ""}
};


static Bool
DialConfig (LocalDevicePtr * array, int inx, int max, LexPtr val)
{
  LocalDevicePtr dev = array[inx];
  DialPtr dial = (DialPtr) (dev->private);
  int token;
  int mtoken;

  xf86ErrorFVerb (1, ErrorF ("DialConfig\n"));

  while ((token = xf86GetToken (DialTab)) != ENDSUBSECTION)
    {
      switch (token)
	{
	case PORT:
	  if (xf86GetToken (NULL) != STRING)
	    xf86ConfigError ("Dial input port expected");
	  dial->port = strdup (val->str);
	  if (xf86Verbose)
	    ErrorF ("%s Dial input port: %s\n", XCONFIG_GIVEN, dial->port);
	  break;

	case DEVICENAME:
	  if (xf86GetToken (NULL) != STRING)
	    xf86ConfigError ("Option string expected");
	  dev->name = strdup (val->str);
	  if (xf86Verbose)
	    ErrorF ("%s Dial X device name is %s\n", XCONFIG_GIVEN,
		    dev->name);
	  break;

	case THE_MODE:
	  mtoken = xf86GetToken (ModeTabRec);
	  if ((mtoken == EOF) || (mtoken == STRING) || (mtoken == NUMBER))
	    xf86ConfigError ("Mode type token expected");
	  else
	    {
	      switch (mtoken)
		{
		case ABSOLUTE:
		  dial->flags = dial->flags | DIAL_FLAG_ABSOLUTE;
		  break;
		case RELATIVE:
		  dial->flags = dial->flags & ~DIAL_FLAG_ABSOLUTE;
		  break;
		default:
		  xf86ConfigError ("Illegal Mode type");
		  break;
		}
	    }
	  break;

	case DEBUG_LEVEL:
	  if (xf86GetToken (NULL) != NUMBER)
	    xf86ConfigError ("Option number expected");
	  debug_level = val->num;
	  if (xf86Verbose)
	    {
#if DEBUG
	      ErrorF ("%s Dial debug level sets to %d\n", XCONFIG_GIVEN,
		      debug_level);
#else
	      ErrorF ("%s Dial debug level not sets to %d because"
		      " debugging is not compiled\n", XCONFIG_GIVEN,
		      debug_level);
#endif
	    }
	  break;

	case HISTORY_SIZE:
	  if (xf86GetToken (NULL) != NUMBER)
	    xf86ConfigError ("Option number expected");
	  dev->history_size = val->num;
	  if (xf86Verbose)
	    ErrorF ("%s Dial Motion history size is %d\n", XCONFIG_GIVEN,
		    dev->history_size);
	  break;

	case ALWAYS_CORE:
	  xf86AlwaysCore (dev, TRUE);
	  if (xf86Verbose)
	    ErrorF ("%s Dial device always stays core pointer\n",
		    XCONFIG_GIVEN);
	  break;

	case EOF:
	  FatalError ("Unexpected EOF (missing EndSubSection)");
	  break;

	default:
	  xf86ConfigError ("Dial subsection keyword expected");
	  break;
	}
    }

  xf86ErrorFVerb (1, ErrorF ("DialConfig name=%s\n", dev->name));

  return Success;
}
#endif

/*
 ***************************************************************************
 *
 * DialConvert --
 *	Convert valuators to X and Y.
 *
 ***************************************************************************
 */
static Bool
DialConvert (LocalDevicePtr local,
	     int first,
	     int num,
	     int v0, int v1, int v2, int v3, int v4, int v5, int *x, int *y)
{
  /*DialPtr dial = (DialPtr) local->private; */

  if (first != 0 || num == 1)
    return FALSE;

  *x = ((v0 + 32768) * screenInfo.screens[0]->width) / 65535;
  *y = ((v1 + 32768) * screenInfo.screens[0]->height) / 65535;

  xf86ErrorFVerb (5, "Dial converted v0=%d v1=%d to x=%d y=%d\n",
		  v0, v1, *x, *y);

  return TRUE;
}

/*
 ***************************************************************************
 *
 * DialReverseConvert --
 *	Convert X and Y to valuators.
 *
 ***************************************************************************
 */
static Bool
DialReverseConvert (LocalDevicePtr local, int x, int y, int *valuators)
{
  /*DialPtr dial = (DialPtr) local->private; */

  valuators[0] = ((x * 65535) / screenInfo.screens[0]->width) - 32768;
  valuators[1] = ((y * 65535) / screenInfo.screens[0]->height) - 32768;

  xf86ErrorFVerb (5, "Dial converted x=%d y=%d to v0=%d v1=%d\n", x, y,
		  valuators[0], valuators[1]);

  return TRUE;
}

/*
 ***************************************************************************
 *
 * DialReadInput --
 *	Read the new events from the device, and enqueue them.
 *
 ***************************************************************************
 */
static void
DialReadInput (LocalDevicePtr local)
{
  DialPtr dial = (DialPtr) local->private;
  int count = 0;
  int ch = 0;
  int is_absolute = (dial->flags & DIAL_FLAG_ABSOLUTE);

  xf86ErrorFVerb (7, "DialReadInput BEGIN fd=%d\n", local->fd);

  /* Limit total input to 200 bytes per call,
   * so we don't hang the server if something goes wrong.
   */
  while (count < 200 && (ch = XisbRead (dial->buffer)) >= 0)
    {
      if ((dial->state != 0) || IS_DIAL_EVENT (ch))
	{
	  switch (dial->state)
	    {
	    case DIAL_WHICH_DEVICE:
	      dial->which = ch - DIAL_BASE;
	      dial->state++;
	      break;
	    case DIAL_VALUE_HIGH:
	      dial->value = (ch << 8);
	      dial->state++;
	      break;
	    case DIAL_VALUE_LOW:
	      {
		int delta;

		dial->value |= ch;

		xf86ErrorFVerb (10, "DialReadInput Valuator %d set to %d\n",
				dial->which, dial->value);

		delta = dial->value - dial->hwvstate[(int) dial->which];
		dial->hwvstate[(int) dial->which] = dial->value;

		xf86PostMotionEvent (local->dev, is_absolute,
				     dial->which, 1,
				     is_absolute ? dial->value : delta);
		dial->state = 0;
		break;
	      }
	    default:
	      ErrorF ("Impossible state %d in dial_intr.\n", dial->state);
	      dial->state = 0;
	      break;
	    }
	  dial->errorCount = 0;
	}
      else if (IS_KEY_PRESS (ch) && (dial->flags & DIAL_FLAG_ENABLE_BUTTONS))
	{
	  unsigned char btnNum = ((ch - DIAL_PRESS_BASE) ^ 0x18) + 1;

	  xf86PostButtonEvent (local->dev, is_absolute,	/* is_absolute */
			       btnNum, 1,
			       0, 8,
			       dial->hwvstate[0],
			       dial->hwvstate[1],
			       dial->hwvstate[2],
			       dial->hwvstate[3],
			       dial->hwvstate[4],
			       dial->hwvstate[5],
			       dial->hwvstate[6], dial->hwvstate[7]);
	  dial->errorCount = 0;
	}
      else if (IS_KEY_RELEASE (ch) && (dial->flags & DIAL_FLAG_ENABLE_BUTTONS))
	{
	  unsigned char btnNum = ((ch - DIAL_RELEASE_BASE) ^ 0x18) + 1;

	  xf86PostButtonEvent (local->dev, is_absolute,	/* is_absolute */
			       btnNum, 0,
			       0, 8,
			       dial->hwvstate[0],
			       dial->hwvstate[1],
			       dial->hwvstate[2],
			       dial->hwvstate[3],
			       dial->hwvstate[4],
			       dial->hwvstate[5],
			       dial->hwvstate[6], dial->hwvstate[7]);
	  dial->errorCount = 0;
	}
      else if (ch == DIAL_INITIALIZED)
	{

	  fprintf (stderr, "Dial box initialized OK.\n");
	  dial_setup (dial);
	  dial->errorCount = 0;
	}
      else
	{
	  ErrorF ("unexpected char %d from dial\n", ch);
	  /* try to reinitialize */
	  if (dial->errorCount++ > 5)
	    {
	      ErrorF ("Reinitializing dial box (not really)\n");
	     /*FIXME*/}
	}
      count++;
    }

  if (ch == -2)
    {
      ErrorF ("Error reading dial device : %s\n", strerror (errno));
      return;
    }
  else
    {
      xf86ErrorFVerb (10, "DialReadInput read %d bytes\n", count);
    }


  xf86ErrorFVerb (7, "DialReadInput END   local=0x%x dial=0x%x\n",
		  local, dial);
}

#if 0
/*
 ***************************************************************************
 *
 * DialLedControlProc --
 *
 ***************************************************************************
 */
static void
DialLedControlProc (DeviceIntPtr dev, LedCtrl * ctrl)
{
  LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
  DialPtr dial = (DialPtr) XI_PRIVATE (dev);

  xf86ErrorFVerb (2,
		  "DialLedControlProc fd=%d led_values=0x%x led_mask=0x%x\n",
		  local->fd, ctrl->led_values, ctrl->led_mask);
  dial_set_led_mask (dial, ctrl->led_values & ctrl->led_mask);
}

/*
 ***************************************************************************
 *
 * DialOpenDevice --
 *	Open the physical device and init information structs.
 * This is opened during the setup process now. Maybe we could
 * make a re-open function here instead.
 ***************************************************************************
 */
static int
DialOpenDevice (DeviceIntPtr dev)
{
  /*
     LocalDevicePtr   local = (LocalDevicePtr)dev->public.devicePrivate;
     DialPtr  dial = (DialPtr)XI_PRIVATE(dev);
     double           screenRatio, tabletRatio;
     int                      gap;

     if (local->fd < 0) {
     DialOpen(local);
     }

     return (local->fd != -1);
   */
  return 0;
}

#endif /* No Config section */

/*
 ***************************************************************************
 *
 * DialCloseProc --
 *
 * The actual device close stuff is in DialTearDownProc
 *
 ***************************************************************************
 */
static void
DialCloseProc (LocalDevicePtr local)
{
}


/*
 ***************************************************************************
 *
 * DialDeviceControl --
 *      Handle the initialization, etc. of a dial
 *
 ***************************************************************************
 */
static int
DialDeviceControl (DeviceIntPtr dev, int what)
{
  CARD8 map[DIAL_NUM_BUTTONS + 1];
  int loop;

  LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
  DialPtr dial = (DialPtr) (local->private);

  xf86ErrorFVerb (2,
		  "BEGIN DialDeviceControl dev=0x%x dial=0x%x flags=%d what=%d\n",
		  dev, dial, dial->flags, what);

  switch (what)
    {
    case DEVICE_INIT:
      xf86ErrorFVerb (1, "DialDeviceControl dev=0x%x what=INIT\n", dev);

      if (1) /*dial->flags & DIAL_FLAG_ENABLE_BUTTONS)*/
	{
	  for (loop = 1; loop <= DIAL_NUM_BUTTONS; loop++)
	    map[loop] = loop;

/* FIXME: If I don't allocat a button class for the dialbox, it and other extension devices get messed up.
	Is it possible that there is a bug somewhere else that expects al valuator devices to have
	at least one button?? For now, I will allocate a  button class with zero buttons, which works OK. */
  if (InitButtonClassDeviceStruct (dev, (dial->flags & DIAL_FLAG_ENABLE_BUTTONS)?DIAL_NUM_BUTTONS:0, map) ==  FALSE)
	    {
	      ErrorF ("unable to allocate Button class device\n");
	      return !Success;
	    }
	}

      if (InitFocusClassDeviceStruct (dev) == FALSE)
	{
	  ErrorF ("unable to init Focus class device\n");
	  return !Success;
	}

/***********************************
 * The Feedbacks don't work yet. It seems that this part of XFree86 4.x
 * is not complete yet anyhow.  -- FIXME
 don't forget that LED events are disabled when buttons are disabled.

      if (InitPtrFeedbackClassDeviceStruct(dev, DialPtrControlProc) == FALSE)
        {
          ErrorF("unable to init ptr feedback\n");
          return !Success;
        }

      if (InitLedFeedbackClassDeviceStruct (dev, DialLedControlProc) == FALSE)
	{
	  ErrorF ("unable to init leds feedback\n");
	  return !Success;
	}

************************************/


      if (InitProximityClassDeviceStruct (dev) == FALSE)
	{
	  ErrorF ("unable to init proximity class device\n");
	  return !Success;
	}

      if (InitValuatorClassDeviceStruct (dev,
					 DIAL_NUM_VALUATORS,
					 xf86GetMotionEvents,
					 local->history_size,
					 ((dial->flags & DIAL_FLAG_ABSOLUTE)
					  ? Absolute : Relative)) == FALSE)
	{
	  ErrorF ("unable to allocate Valuator class device\n");
	  return !Success;
	}
      else
	{
	  /* init ranges of valuators */
	  for (loop = 0; loop < DIAL_NUM_VALUATORS; loop++)
	    {
	      InitValuatorAxisStruct (dev, loop, -32768,	/* min val */
				      32767,	/* max val */
				      DIAL_VALUATOR_RESOLUTION,	/* resolution */
				      0,	/* min_res */
				      400);	/* max_res */

	    }

	  /* allocate the motion history buffer */
	  xf86MotionHistoryAllocate (local);

	  /*AssignTypeAndName (dev, local->atom, local->name); */
	}

      /* open the device to gather informations */
      /*DialOpenDevice (dev); */

      xf86ErrorFVerb (1, "DialDeviceControl INIT complete\n", dev);
      break;



      /* Here we re-open the serial port, which is close while the device is OFF */
    case DEVICE_ON:
      xf86ErrorFVerb (1, "DialDeviceControl dev=0x%x what=ON\n", dev);

      local->fd = xf86OpenSerial (local->options);
      if (local->fd == -1)
	{
	  xf86Msg (X_WARNING, "%s: cannot open input device\n", local->name);
	  return (!Success);
	}

      dial->buffer = XisbNew (local->fd, DIAL_BUFFER_SIZE);
      if (!dial->buffer)
	{
	  xf86CloseSerial (local->fd);
	  local->fd = -1;
	  return (!Success);
	}

      xf86FlushInput (local->fd);
      AddEnabledDevice (local->fd);
      dev->public.on = TRUE;
      break;

    case DEVICE_OFF:
    case DEVICE_CLOSE:
      xf86ErrorFVerb (1, "DialDeviceControl  dev=0x%x what=%s\n", dev,
		      (what == DEVICE_CLOSE) ? "CLOSE" : "OFF");
      /*dial_clear_leds (dial); */

      if (local->fd != -1)
	{
	  RemoveEnabledDevice (local->fd);
	  if (dial->buffer)
	    {
	      XisbFree (dial->buffer);
	      dial->buffer = NULL;
	    }
	  xf86CloseSerial (local->fd);
	}

      RemoveEnabledDevice (local->fd);
      dev->public.on = FALSE;
      break;

    default:
      ErrorF ("Dial: unsupported DeviceControl mode=%d\n", what);
      return !Success;
      break;
    }
  xf86ErrorFVerb (2,
		  "END   DialDeviceControl Success what=%d dev=0x%x dial=0x%x\n",
		  what, dev, dial);
  return Success;
}

/*
 ***************************************************************************
 *
 * DialControlProc --
 *
 ***************************************************************************
 */
static int
DialControlProc (LocalDevicePtr local, xDeviceCtl * control)
{
  xDeviceResolutionCtl *res;
  int *resolutions;
  char str[10];
  DialPtr dial = (DialPtr) local->private;

  res = (xDeviceResolutionCtl *) control;
  if ((control->control != DEVICE_RESOLUTION) || (res->num_valuators < 1))
    return (BadMatch);

  resolutions = (int *) (res + 1);

  xf86ErrorFVerb (3, "DialControlProc changing to %d (suppressing under)\n",
		  resolutions[0]);

  sprintf (str, "SU%d\r", resolutions[0]);
  XisbWrite (dial->buffer, str, strlen (str));

  return (Success);
}

/*
 ***************************************************************************
 *
 * DialSwitchMode --
 *
 ***************************************************************************
 */
static int
DialSwitchMode (ClientPtr client, DeviceIntPtr dev, int mode)
{
  LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
  DialPtr dial = (DialPtr) local->private;

  xf86ErrorFVerb (3, "DialSwitchMode dev=0x%x mode=%d\n", dev, mode);

  if (mode == Absolute)
    {
      dial->flags = dial->flags | DIAL_FLAG_ABSOLUTE;
    }
  else
    {
      if (mode == Relative)
	{
	  dial->flags = dial->flags & ~DIAL_FLAG_ABSOLUTE;
	}
      else
	{
	  xf86ErrorFVerb (1,
			  "DialSwitchMode dev=0x%x invalid mode=%d\n", dev,
			  mode);
	  return BadMatch;
	}
    }
  return Success;
}

#define SYSCALL(call) while(((call) == -1) && (errno == EINTR))
#define CHECKEDWRITE(fd, buff, len) SYSCALL(write((fd), (buff), (len)))

static int
DialQueryHardware (LocalDevicePtr local)
{
  unsigned char dcmd[1];
  int times;
  DialPtr dial = (DialPtr) local->private;
  int len;

  xf86ErrorFVerb (1, "Dial box initialization in progress.....\n");

  dcmd[0] = DIAL_INITIALIZE;

  times = 10;
  do
    {
      unsigned char ch;

      CHECKEDWRITE (local->fd, dcmd, sizeof (dcmd));
      times--;
      wait_for_fd (local->fd);
      SYSCALL (len = read (local->fd, &ch, 1));

      xf86ErrorFVerb (10, "received[%d] 0x%02x\n", times, ch);

      dial->initialized = (len == 1) && (ch == DIAL_INITIALIZED);
    }
  while (times > 0 && !dial->initialized);

  dial_setup (dial);

  ErrorF ("Dial box initialization %s\n",
	  dial->initialized ? "done" : "failed");

  if (len <= 0)
    {
      SYSCALL (close (local->fd));
      local->fd = -1;
      return !Success;
    }

  return Success;
}


#ifndef XFree86LOADER
/*
 ***************************************************************************
 *
 * wait_for_fd --
 *
 *      Wait one second that the file descriptor becomes readable.
 *
 ***************************************************************************
 */
static int
wait_for_fd (int fd)
{
  int err;
  fd_set readfds;
  struct timeval timeout;

  FD_ZERO (&readfds);
  FD_SET (fd, &readfds);

  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  SYSCALL (err = select (FD_SETSIZE, &readfds, NULL, NULL, &timeout));

  return err;
}

#endif
