
#define	DIAL_NUM_BUTTONS		32
#define	DIAL_NUM_LEDS			32

#define	DIAL_NUM_VALUATORS		8
#define	DIAL_VALUATOR_RESOLUTION	256
#define	DIAL_VALUATOR_MIN_FILTER	0
#define	DIAL_VALUATOR_MAX_FILTER	7

#define DIAL_BUFFER_SIZE		256


	/* bit definitions for private flags */
#define DIAL_FLAG_ABSOLUTE       1
#define DIAL_FLAG_ENABLE_BUTTONS 2

	/* dial parser state machine states */
#define DIAL_WHICH_DEVICE       0
#define DIAL_VALUE_HIGH         1
#define DIAL_VALUE_LOW          2
#define DIAL_N_STATES           3

	/* dial/button box commands */
#define DIAL_INITIALIZE                 0x20
#define DIAL_SET_LEDS                   0x75
#define DIAL_SET_TEXT                   0x61
#define DIAL_SET_AUTO_DIALS             0x50
#define DIAL_SET_AUTO_DELTA_DIALS       0x51
#define DIAL_SET_FILTER			0x53
#define DIAL_SET_BUTTONS_MOM_TYPE       0x71
#define DIAL_SET_AUTO_MOM_BUTTONS       0x73
#define DIAL_SET_ALL_LEDS		0x4b
#define DIAL_CLEAR_ALL_LEDS		0x4c

	/* dial/button box replies and events */
#define DIAL_INITIALIZED        0x20
#define DIAL_BASE               0x30
#define DIAL_DELTA_BASE         0x40
#define DIAL_PRESS_BASE         0xc0
#define DIAL_RELEASE_BASE       0xe0

	/* macros to determine reply type */
#define IS_DIAL_EVENT(ch)       (((ch)>=DIAL_BASE)&&((ch)<DIAL_BASE+DIAL_NUM_VALUATORS))
#define IS_KEY_PRESS(ch)        (((ch)>=DIAL_PRESS_BASE)&&((ch)<DIAL_PRESS_BASE+DIAL_NUM_BUTTONS))
#define IS_KEY_RELEASE(ch)      (((ch)>=DIAL_RELEASE_BASE)&&((ch)<DIAL_RELEASE_BASE+DIAL_NUM_BUTTONS))
#define IS_INIT_EVENT(ch)       ((ch)==DIAL_INITIALIZED)

/* mask manipulation */
#define NBITS 8

#define SET(n, p)   ((p)[(n)/NBITS] |= ((unsigned)1 << ((n) % NBITS)))

#define CLR(n, p)   ((p)[(n)/NBITS] &= ~((unsigned)1 << ((n) % NBITS)))

#define ISSET(n, p) ((p)[(n)/NBITS] & ((unsigned)1 << ((n) % NBITS)))

/* Private data structure */

typedef struct _DialRec
{
  XISBuffer *buffer;
  int flags;

  /* Button and valuators states */
  unsigned char vactive[DIAL_NUM_VALUATORS];
  int hwvstate[DIAL_NUM_VALUATORS];
  unsigned char bactive[DIAL_NUM_BUTTONS];
  unsigned char bstate[DIAL_NUM_BUTTONS];
  unsigned char ledstate[DIAL_NUM_LEDS];

  /* state machine stuff */
  char state;
  char which;
  short value;
  void *checkinit_id;

  /* Misc. Std. Driver stuff */
  unsigned char initialized;	/* true if data structures are set */
  /* and init sequence has been sent to */
  /* device */

  unsigned char ready;		/* true if we've received confirmation */
  /* of the init sequence from the device */

  unsigned char errorCount;	/* count of bad messages for recovery */
}
DialRec, *DialPtr;

/* Internal functions */
static int dial_enable_button (DialRec * dial, int btns, int value);
static int dial_enable_valuator (DialRec * dial, int val, int value);
static int dial_set_resolution (DialRec * dial, int valNum, int valFilter);
#if 0
static int dial_switch_leds (DialRec * dial);
static int dial_clear_leds (DialRec * dial);
static int dial_set_led (DialRec * dial, int led, int value);
static int dial_set_led_mask (DialRec * dial, unsigned int mask);
#endif
static void dial_setup (DialRec * dial);

/* Exported functions */
static InputInfoPtr DialPreInit (InputDriverPtr drv, IDevPtr dev, int flags);
static pointer DialSetupProc (pointer module, pointer options,
			      int *errmaj, int *errmin);
static void DialTearDownProc (pointer p);

	/* These go into the LocalDeviceRec */
static Bool DialDeviceControl (DeviceIntPtr dev, int what);
static void DialReadInput (LocalDevicePtr local);
static int DialControlProc (LocalDevicePtr local, xDeviceCtl * control);
static void DialCloseProc (LocalDevicePtr local);
static int DialSwitchMode (ClientPtr client, DeviceIntPtr dev, int mode);
static Bool DialConvert (LocalDevicePtr local, int first, int num, int v0,
			 int v1, int v2, int v3, int v4, int v5, int *x,
			 int *y);
static Bool DialReverseConvert (LocalDevicePtr local, int x, int y,
				int *valuators);


/*static void DialLedControlProc (DeviceIntPtr dev, LedCtrl * ctrl);*/
/*static int DialOpenDevice (DeviceIntPtr dev);*/
static int DialQueryHardware (LocalDevicePtr local);

#ifndef XFree86LOADER
static int wait_for_fd (int fd);
#endif
